#' @title BgeeMetadata S4 class
#' 
#' @description An S4 class that contains all information to retrieve 
#' intergenic regions generated by Bgee.
#' 
#' @slot intergenic_release Bgee intergenic release that will be used
#' @slot intergenic_prefix String used to generate an intergenic release 
#' specific output directory
#' @slot fasta_intergenic_url URL to fasta file containing reference 
#' intergenic regions as defined by the Bgee pipeline 
#' (https://github.com/BgeeDB/bgee_pipeline/tree/master/pipeline/RNA_Seq)
#' @slot fasta_intergenic_name Name of the fasta file containing all 
#' intergenic regions that will be download to get_species_path()
#' @slot ref_intergenic_ids_name Name of the file containing the list of true 
#' intergenic regions that will be download to get_species_path()
#'
BgeeMetadata <- setClass(
    # Set the name for the class
    Class = "BgeeMetadata",
    
    # Define the slots
    representation = representation(
        intergenic_release = "character",
        intergenic_prefix = "character",
        fasta_intergenic_url = "character",
        fasta_intergenic_name = "character",
        ref_intergenic_ids_name = "character"
        
    ),
    
    prototype = prototype(
        intergenic_prefix = "intergenic_",
        fasta_intergenic_name = "bgee_intergenic.fa.gz",
        ref_intergenic_ids_name = "bgee_ref_intergenic.tsv"
    )
)

setMethod(f = "initialize", signature ="BgeeMetadata" , 
          definition = function(.Object, ...) {
              .Object <- callNextMethod()
              ## Get release information
              cat("\nQuerying Bgee to get intergenic release information...\n")
              allReleases <- try(.getIntergenicRelease(removeFile = FALSE), 
                                 silent=TRUE)
              if ( is(allReleases, "data.frame") ){
                  file.rename(from=file.path(getwd(), 'release.tsv'),
                              to=file.path(getwd(), "release.tsv"))
              } else if ( is(allReleases, "try-error") ){
                  if (file.exists(file.path(getwd(), "release.tsv"))){
                      cat(paste0("\nWARNING: BgeeCall could not access 
intergenic releases information from the internet, but a release 
information file was found in the download directory ",
getwd(), ". This release file will be used, but be warned that it
may not be up to date!\n"))
                      allReleases <- read.table(file.path(getwd(), 
                                                          "release.tsv"), 
                                                header=TRUE, sep="\t")
                  } else {
                      stop("ERROR: BgeeCall could not access Bgee 
intergenic releases information. Is your internet connection working?")
                  }
              }
              
              # if no release has been specified during instanciation 
              # take more recent release
              if (length(.Object@intergenic_release) == 0) {
                  .Object@intergenic_release <- 
                      as.character(allReleases$release[1])
                  .Object@fasta_intergenic_url <- 
                      as.character(allReleases$referenceIntergenicFastaURL[1])
                  message_to_users <- 
                      as.character(allReleases$messageToUsers[1])
              } else if (length(.Object@intergenic_release) == 1) {
                  if (.Object@intergenic_release %in% allReleases$release) {
                      .Object@intergenic_release <- 
                          as.character(allReleases$release[allReleases$release == 
                                                               .Object@intergenic_release])
                      .Object@fasta_intergenic_url <- 
                          as.character(allReleases$referenceIntergenicFastaURL[allReleases$release == 
                                                                                   .Object@intergenic_release])
                      message_to_users <- 
                          as.character(allReleases$messageToUsers[allReleases$release == 
                                                                      .Object@intergenic_release])
                  } else {
                      stop("ERROR: The specified release number is invalid, 
or is not available for BgeeCall.")
                  }
              } else {
                  stop("ERROR: The specified release number is invalid.")
              }
              if(message_to_users != "") {
                  message("IMPORTANT : ", message_to_users)
              }
              validObject(.Object)
              .Object
          })

#' @title `intergenic_release` Setter
#' 
#' @description Set value of the `intergenic_release` slot
#' 
#' @param bgeeObject The BgeeMetadata object
#' @param intergenicRelease character corresponding to the `intergenic_release`
#' 
#' @return An object of the class BgeeMetadata with new `intergenic_release`
#'  value
#' 
#' @export
#' @docType methods
#' @rdname setIntergenicRelease
#' 
#' @examples {
#' bgee <- new("BgeeMetadata")
#' bgee <- setIntergenicRelease(bgee, "0.1")
#' }
#'
setGeneric(name="setIntergenicRelease", 
           def=function(bgeeObject, intergenicRelease) {
               standardGeneric("setIntergenicRelease")
           })

#' @title `intergenic_release` Getter
#' 
#' @description Get value of the `intergenic_release` slot
#' 
#' @param bgeeObject The BgeeMetadata object
#' 
#' @return the value of the `intergenic_release` slot of the object
#' 
#' @export
#' @docType methods
#' @rdname getIntergenicRelease
#' 
#' @examples {
#' bgee <- new("BgeeMetadata")
#' intergenic_release <- getIntergenicRelease(bgee)
#' }
#'
setGeneric(name="getIntergenicRelease", def=function(bgeeObject) {
    standardGeneric("getIntergenicRelease")
})

#' @title `intergenic_prefix` Getter
#' 
#' @description Get value of the `intergenic_prefix` slot
#' 
#' @param bgeeObject The BgeeMetadata object
#' 
#' @return the value of the `intergenic_prefix` slot of the object
#' 
#' @export
#' @docType methods
#' @rdname getIntergenicPrefix
#' 
#' @examples {
#' bgee <- new("BgeeMetadata")
#' intergenic_prefix <- getIntergenicPrefix(bgee)
#' }
#'
setGeneric(name="getIntergenicPrefix", def=function(bgeeObject) {
    standardGeneric("getIntergenicPrefix")
})

#' @rdname setIntergenicRelease
#' @aliases setIntergenicRelease,bgeeMetadata,character
setMethod(f="setIntergenicRelease",
          signature=c(bgeeObject = "BgeeMetadata", 
                      intergenicRelease = "character"), 
          definition=function(bgeeObject, intergenicRelease) {
              bgeeObject@intergenic_release <- intergenicRelease
              return(bgeeObject)
          })

#' @rdname getIntergenicRelease
#' @aliases getIntergenicRelease,bgeeMetadata
setMethod(f="getIntergenicRelease", 
          signature=c(bgeeObject = "BgeeMetadata"), 
          definition=function(bgeeObject) {
              return(bgeeObject@intergenic_release)
          })

#' @rdname getIntergenicPrefix
#' @aliases getIntergenicPrefix,bgeeMetadata
setMethod(f="getIntergenicPrefix", 
          signature=c(bgeeObject = "BgeeMetadata"), 
          definition=function(bgeeObject) {
              return(bgeeObject@intergenic_prefix)
          })
